/** 
 * Add <i>listener</i> to <i>bag</i> of listeners. The function does not modify <i>bag</i> and return a new collection containing <i>listener</i> and all listeners from <i>bag</i>. Bag without listeners always represented as the null value. <p> Usage example: <pre> private volatile Object changeListeners; public void addMyListener(PropertyChangeListener l) { synchronized (this) { changeListeners = Kit.addListener(changeListeners, l); } } public void removeTextListener(PropertyChangeListener l) { synchronized (this) { changeListeners = Kit.removeListener(changeListeners, l); } } public void fireChangeEvent(Object oldValue, Object newValue) { // Get immune local copy Object listeners = changeListeners; if (listeners != null) { PropertyChangeEvent e = new PropertyChangeEvent( this, "someProperty" oldValue, newValue); for (int i = 0; ; ++i) { Object l = Kit.getListener(listeners, i); if (l == null) break; ((PropertyChangeListener)l).propertyChange(e); } } } </pre>
 * @param listener Listener to add to <i>bag</i>
 * @param bag Current collection of listeners.
 * @return A new bag containing all listeners from <i>bag</i> and<i>listener</i>.
 * @see #removeListener(Object bag,Object listener)
 * @see #getListener(Object bag,int index)
 */
public static Object addListener(Object bag,Object listener){
  if (listener == null)   throw new IllegalArgumentException();
  if (listener instanceof Object[])   throw new IllegalArgumentException();
  if (bag == null) {
    bag=listener;
  }
 else   if (!(bag instanceof Object[])) {
    bag=new Object[]{bag,listener};
  }
 else {
    Object[] array=(Object[])bag;
    int L=array.length;
    if (L < 2)     throw new IllegalArgumentException();
    Object[] tmp=new Object[L + 1];
    System.arraycopy(array,0,tmp,0,L);
    tmp[L]=listener;
    bag=tmp;
  }
  return bag;
}
