#!/usr/bin/ruby
require 'json'
require 'time'
require 'fileutils'


KEYS = ['fix', 'fixed']
P_CODE = "LUCENE"

INPUT_LOG = ".\/lucene_log.txt"
OUT_REF_LIST = ".\/lucene-ref.txt"
OUT_REV_JSON_DATA = ".\/lucene-rev.json"

INPUT_REPORTS_DIR = ".\/bug_report_data"
OUT_REPORTS_JSON = ".\/lucene-reports.json"

FINAL_BUGS_JSON = ".\/lucene-bugs.json"
  
module Helper

  def Helper.extract_bugs(bug_list_dir = "/Users/assist/dataset/TCM/DLo/amldata/ground_truth/lucene")
    files = Dir.entries(bug_list_dir).select do |bug_id|
      ! bug_id.start_with?(".")      
    end

    bugs = []
    files.each do |f|
      bugs << f.split(".")[0]      
    end   
    bugs
  end
      
  def Helper.filter(bug)
    (Helper.constraints_on_commit_message(bug) && Helper.constraints_on_bug_id(bug) && Helper.constraints_on_changed_files(bug)) ? true : false
  end

  def Helper.constraints_on_bug_id(bug) # only revision that deals with more than 1 bug is considered
    return false if bug[:ref].nil?
    bug[:ref].size >= 1 ? true : false
  end

  def Helper.constraints_on_1_bug_id(bug) # only revision that deals with 1 bug is considered
    return false if bug[:ref].nil?
    bug[:ref].size == 1 ? true : false
  end

  
  def Helper.constraints_on_commit_message(bug, keys = KEYS) # commit log message must contain some key words
    m = bug[:message]
    return false if m.nil?
    keys.each do |k|
      m.match(/\s*(#{k})\s+/i) ? (return true) : (return false)
    end
  end
  
  def Helper.constraints_on_changed_files(bug)
    return false if bug[:changed_files].nil?    
    bug[:changed_files].each do |key, value|
      return false if ! (key.end_with?(".java") || key.end_with?(".txt")) # all files in .java or .txt are considered
      return false if ! "A M".match(value.strip.upcase) # all Added or Modified files are consideree
      if value.strip.upcase.eql?("A")
        return false if ! key.strip.match(/.test/i) # if a file is added, it must be a test file to be considered
      end
    end
    return true
  end
  
  def  Helper.changed_files_with_only_one_file(bug)
    return false if bug[:changed_files].nil?
    bug[:changed_files].size == 1 ? true : false
  end

  def Helper.filter_with_test(json_data)
    bugs_with_test = []
    json_bugs = {}
    json_data.each do |key, value|      
      value.each do |bug|
        bug["changed_files"].each do |file, status|
          if file.strip.match(/.test/)
            bugs_with_test << bug
            break
          end
        end
      end
      json_bugs = Hash[key => bugs_with_test]
      break
    end

    puts "[bugs with tests either added or modifiled are #{bugs_with_test.size}] ... [ok]"
    json_bugs
  end

  def Helper.elim_revs_deal_with_same_bug_id(bugs_json, project = P_CODE)
    bugs_json[project].each_with_index do |bug, index|
      bug[:c_ref] = []
      bugs_json[project].each_with_index do |b, i|
        bug[:c_ref].concat b[:ref] if index != i
      end
    end
    bugs_json[project].delete_if do |bug|
      x = bug[:ref] &  bug[:c_ref]
      ! x.empty?
    end

    bugs_json[project].each do |bug|
      bug.delete(:c_ref)      
    end
    puts "[Unique revisions, i.e., non two revisions fix the same bug] ... [#{bugs_json[project].size}] ... [ok]"
    bugs_json
  end


  def Helper.apply_filter(json_bugs, bug_with_ref_id = true, unique_revisions = true, revision_fix_exactly_1_bug = true, key_words = true, changed_files = true)
    if bug_with_ref_id
      json_bugs["#{P_CODE}"].delete_if {|bug| ! Helper.constraints_on_bug_id(bug)} 
      puts "[Revisions that have bug report reference] ... [#{json_bugs[P_CODE].size}] ... [ok]"
    end
    if unique_revisions
      Helper.elim_revs_deal_with_same_bug_id(json_bugs)
    end

    if revision_fix_exactly_1_bug
      json_bugs[P_CODE].delete_if do |bug|
        ! Helper.constraints_on_1_bug_id(bug)
      end
      puts "[Revisions that fix exactly only 1 bug] ... [#{json_bugs[P_CODE].size}] ... [ok]"
    end

    if key_words
      json_bugs[P_CODE].delete_if do |bug|
        ! Helper.constraints_on_commit_message(bug)
      end
      puts "[Revisions that log message explicitly mentioned: fix or fixed] ... [#{json_bugs[P_CODE].size}] ... [ok]"
    end

    if changed_files
      json_bugs[P_CODE].delete_if do |bug|
        ! Helper.constraints_on_changed_files(bug)
      end
      puts "[Revisions that modified or added .java files only] ... [#{json_bugs[P_CODE].size}] ... [ok]"
    end

    json_bugs
    
  end

  def Helper.extract_to_json(input_file = INPUT_LOG, project_name = P_CODE)    
    dem = "------------------------------------------------------------------------"
    f = File.open(input_file, "r")
    logs = f.read().strip
    f.close
    log_array= logs.split(dem).reject {|e| e.empty?}
    bugs = []
    revs = []
    prev = nil
    
    log_array.each do |log|
      next if log.split("\n\n").empty?
      log_head = log.split("\n\n")[0].strip
      log_message = log.split("\n\n")[1].strip
      if project_name.eql?("LUCENE")
        ref = log_message.scan(/#{project_name}-\d+/i)
      elsif project_name.eql?("ANT")        
        m = log_message.match(/(PR|bug|Bugzilla|issue|report|bugzilla id)([:\s,#]*\d+)+/i)
        ref = m.to_s.scan(/\d+/).uniq
        ref.map! {|r| "ANT-" + r}
      else
        puts "This project is not supported!!! return now"
        return -1
      end
      fields = log_head.split("Changed paths:")
      header = fields[0].split("|")  
      rev = header[0].strip
      revs << rev      
      author = header[1].strip
      date = header[2].strip
      changed_files = {}
      fields[1].split("\n").reject {|cf| cf.empty?}.each do |file|
        parts = file.strip.split(" ")
        f_status = parts[0].strip
        f_name = parts[1].strip.gsub("\/", ".")[1..-1]
        changed_files[f_name] = f_status
      end
      
      bug = {:rev => rev, :author => author, :date => date, :ref => ref, :message => log_message, :changed_files => changed_files}           
      bugs << bug
    end

    revs.sort! {|x,y| y <=> x}
    bugs.each_with_index do |bug, index|      
      bugs[index][:prev] = revs[revs.index(bug[:rev])+1]
    end

    puts "[Revisions have been extracted into a json data format] ... [#{bugs.size}] ... [ok]"
    Hash[project_name => bugs]
  end

  def Helper.write_json_to_file(json, output_file = OUT_REV_JSON_DATA)
    File.open(output_file, "w") do |f|
      f.write(json.to_json)
    end
    puts "[json data has been successfully saved to #{output_file}] ... [ok]"
  end

  def Helper.extract_bugs_id(bug_json, output = OUT_REF_LIST)
    if bug_json.size != 1
      puts "Something wrong! json data size is expected to be 1."
      return -1
    end    
    bug_json.each do |key, value|      
      File.open(output, "w") do |f|
        value.each do |b|
          f.puts("#{b[:rev]} : #{b[:ref][0].strip}")
        end
      end
    end
    puts "[Bug report number list has been successfully saved to #{OUT_REF_LIST}] ... [ok]"
  end

  def Helper.bug_reports_to_json(input_dir = INPUT_REPORTS_DIR, output_json_file = OUT_REPORTS_JSON)
    bugs = []
    Dir.entries(input_dir).select {|f| !f.start_with?(".")}.each do |f|
      rev = f.split("_")[0]
      ref = f.split("_")[1]
      elem = {}
      puts f
      f = File.open(input_dir + "/" + f, "r")
      hash = {}
      f.each_line do |line|
        hash[line.split(":")[0].strip] = line.split(":")[1].strip
      end
      f.close
      elem[:rev] = rev
      elem[:ref] = ref
      elem[:report] = hash
      bugs << elem
    end
    reports_data = {}
    reports_data[:lucene_reports] = bugs
    File.open(output_json_file, "w") do |f|
      f.write(reports_data.to_json)
    end
    puts "[Bug reports data have been saved to #{output_json_file}] ... [ok]"    
  end

  def Helper.is_bug_and_fixed(bug)
    report = bug["bug_report"]
    return false if report.nil?
    (report["Type"].eql?("Bug") && report["Status"].eql?("Fixed")) ? true : false
  end

  def Helper.parse_json_file(file_path)
    file = File.read(file_path)
    data = JSON.parse(file)
    #puts data
    return data
  end

  def Helper.merge_filter_save(input_rev_json = OUT_REV_JSON_DATA, input_report_json = OUT_REPORTS_JSON, out_final_bug_json = FINAL_BUGS_JSON)
    rev_json_file = File.read(input_rev_json)
    rev_json = JSON.parse(rev_json_file)
    report_json_file = File.read(input_report_json)
    report_json = JSON.parse(report_json_file)

    rev_array = []
    rep_array = []
    rev_json.each do |key, value|
      rev_array = value
    end

    report_json.each do |key, value|
      rep_array = value
    end

    if rev_array.size != rep_array.size
      puts "Something wrong! rev data size must be equal to that of reports size."
    end

    bugs= []
    json_bugs = {}
    rev_array.each do |bug|
      rev = bug["rev"]
      rep_array.each do |rep|
        if rev.eql?(rep["rev"])
          bug["bug_report"] = rep["report"]
          bugs << bug
          break
        end
      end
    end

    filtered_bugs = []
    bugs.each do |bug|
      filtered_bugs << bug if Helper.is_bug_and_fixed(bug)
    end
    
    json_bugs["#{P_CODE}"] = filtered_bugs
    Helper.write_json_to_file(json_bugs, out_final_bug_json)
    puts "[Final bugs list is saved into a json format at #{out_final_bug_json}] ... [Total bugs: #{filtered_bugs.size}] "
  end

  def Helper.extract_rev_prev(bug_json_file = FINAL_BUGS_JSON, output = ".\/rev-prev")
    data = Helper.parse_json_file(bug_json_file)
    file = File.open(output, "w")
    data.each do |key, value|
      value.each do |bug|
        rev = bug["rev"]
        prev = bug["prev"]
        file.puts("#{rev} : #{prev}")
      end
    end
    file.close
    puts "[fixed and bugy revisions numbers have been saved to: #{output}] ... [ok]"
  end

  def Helper.select(bugs_json_file, prev_list_file, bugs_json_file_out)
    bugs = Helper.parse_json_file(bugs_json_file)["LUCENE"]
    prevs = File.open(prev_list_file, "r").read().split("\n")    
    bugs.delete_if do |bug|
      ! prevs.include?(bug["prev"][1..-1])
    end
    Helper.write_json_to_file(Hash["LUCENE" => bugs], bugs_json_file_out)
  end

  def Helper.print_bug(bug, no)
    display_string = ""
    puts "--------------Bug No.#{no+1} -----------------------"
    time = Helper.s_to_d Time.parse(bug["bug_report"]["Resolved"]) - Time.parse(bug["bug_report"]["Created"])
    display_string += "* Bug: #{bug["ref"][0]} \n [-Type: #{bug["bug_report"]["Type"]} -Status: #{bug["bug_report"]["Status"]} -Reporter: #{bug["bug_report"]["Reporter"]} -Assignee: #{bug["bug_report"]["Assignee"]} -Time spent: #{time}] \n"
    display_string += "- Report: Title => #{bug["bug_report"]["Title"]} \n- Description => #{bug["bug_report"]["Description"]}"
    display_string +=       "\n"
    display_string +=       "-Commit message: #{bug["message"]} \n"
    display_string += "\n"
    display_string +=       "- Changed files [#{bug["changed_files"].size}]: \n"
    bug["changed_files"].each do |file, status|
      display_string +=         "#{status} #{file} \n"
    end
    display_string +=       "\n"
    display_string +=       "- faulty version: #{bug["prev"]} => fixed version: #{bug["rev"]} \n"
    display_string += "\n"
    display_string += "- diff: \n #{bug["diff"]}"
    puts display_string
    display_string
  end

  def Helper.print_bugs(bugs_file)
    Helper.parse_json_file(bugs_file)["LUCENE"].each_with_index do |bug, index|
      Helper.print_bug(bug,index)
      puts "\n"
    end
  end

  def Helper.s_to_d(seconds)   
    mm, ss = seconds.divmod(60)   
    hh, mm = mm.divmod(60)  
    dd, hh = hh.divmod(24)  
    return "%d days, %d hours" % [dd, hh]
  end

  def Helper.build_json_bugs_to_files(json_bugs_file = ".\/lucene-bugs-w-clover.json", dest_dir = "/Users/assist/dataset/TCM/dmz/#{P_CODE}")
    FileUtils.rm_rf(Dir.glob("#{dest_dir}/*"))
    bugs = Helper.parse_json_file(json_bugs_file)[P_CODE]
    bugs.each_with_index do |bug, index|
      ref = bug["ref"][0]
      name = "#{dest_dir}/#{ref}"
      if ! File.file?(name) 
        f = File.open(name, "w")
        f.puts(Helper.print_bug(bug, index))
        f.close
      else
        puts "Something wrong!"
      end
    end
  end

  def Helper.create_bugs_map(input = ".\/lucene-bugs-w-clover.json", output = "/Users/assist/ndataset/dlo/dmz/bug_maps/#{P_CODE}")
    f = File.open(output, "w")
    Helper.parse_json_file(input)[P_CODE].each do |bug|
      rev = bug["rev"]
      prev = bug["prev"]
      ref = bug["ref"][0]
      f.puts("#{ref} : #{prev} : #{rev}")
    end
    f.close
    puts "Bugs map has been saved to file: #{output}"
  end 

  def Helper.add_diff(json_bugs_file = ".\/lucene-bugs-w-clover.json", diff_dir = ".\/diff", out_file = ".\/lucene-bugs-w-clover-w-diff.json")
    files = Dir.entries(diff_dir).select do |bug_id|
      ! bug_id.start_with?(".")      
    end

    hash = Helper.parse_json_file(json_bugs_file)
    json_bugs = hash[P_CODE]
    
    files.each do |file|
      if  file.nil?
        puts "something wrong"
        next
      end
      f = File.open("#{diff_dir}/#{file}")
      diff = f.read()
      f.close
      ref = file.split(".")[0].strip
      found = false
      json_bugs.each do |bug|
        if bug["ref"][0].strip.eql?(ref)
          bug[:diff] = diff
          found = true
          break
        end
      end
      if ! found
        puts "something wrong"
        return -1
      else
        found = false
      end
    end
    Helper.write_json_to_file(hash, out_file)
    puts "[Json data has been saved to #{out_file}] ... [ok]"
  end

  def Helper.has_commons(array_1, array_2)
    if array_1.nil? || array_2.nil? || array_1.empty? || array_2.empty?
      return false
    end
    ! (array_1 & array_2).empty?    
  end

  def Helper.a_contain_b(a, b)
    if a.nil? || b.nil? || a.empty? || b.empty?
      return false
    end

    b.each do |e|
      return false if ! a.include?(e.strip)
    end

    return true
  end
  

  def Helper.get_bugs_by_refs(json_bugs, ref_list, project = P_CODE)
    json_bugs[project].keep_if{|bug| Helper.a_contain_b(ref_list,bug[:ref])}
    json_bugs
  end

  def Helper.extract_failed_tests(log_dir, out_dir)
    files = Dir.entries(log_dir).select do |bug_id|
      ! bug_id.start_with?(".")      
    end

    files.each do |file|
      f = File.open("#{out_dir}/#{file}", "w")
      File.readlines("#{log_dir}/#{file}").each do |line|
        test_class = line.split(",")[0]
        test_method = line.split(",")[1]
        status = line.split(",")[2]
        #puts status
        if status.strip.eql? "failed"          
          f.puts("#{test_class}:#{test_method}") 
        end

      end
      f.close()
    end
  end

  def Helper.preprocess(ant_log_file, ant_bug_list, out_file, project)
    bugs_list = Helper.extract_bugs(ant_bug_list) # list of bug_ids
    bugs_json_all = Helper.extract_to_json(ant_log_file, project) # hash of all bugs with all detail data
    revs = Helper.get_bugs_by_refs(bugs_json_all, bugs_list, project)    
    puts "Number of bugs: #{revs[project].size}"
    #data = Helper.elim_revs_deal_with_same_bug_id(revs, project)
    #puts "Number of qualified bugs: #{data[project].size}"
    f = File.open(out_file, "w")
    revs[project].each do |bug|
      f.puts(("#{bug[:ref][0]} : #{bug[:prev]} : #{bug[:rev]}"))
    end
    f.close
  end

  
  ############################################################################
  #
  # Convert coverage data at method level to class level, preprocess method
  # coverage for filtering and indexing source code
  #
  ############################################################################

  def Helper.raw_mcv_to_json(input_file)
#    tests = []
    traces = {}
    test = {}
    test_name = ""
    test_result = ""
    hash = {}
    File.readlines(input_file).each do |line|      
      next if ! line.strip.start_with?("[java]")    # ignore header file
      temp = line.strip.split("\s")[1]
      
      if temp.strip.start_with?("<Test>")

        if ! traces.empty?
          test[:result] = test_result
          test[:traces] = traces
          hash.merge!(test_name => test)
          traces = {}
          test = {}
        end
        
        test_name = temp.strip.split("-")[0][6..-1].strip
        test_result = temp.strip.split("-")[1].eql?("true")? "passed" : "failed"
        next
      end
      
      method = temp.strip.split("-")[0]
      visited_times = temp.strip.split("-")[1]
      traces[method] = visited_times    
     
    end

    test[:result] = test_result
    test[:traces] = traces
    
    hash.merge!(test_name => test)
  end
 
  
  def Helper.get_method_coverage_data(input_file)
    tests = {}
    traces = {}
    File.readlines(input_file).each do |line|      
      next if ! line.strip.start_with?("[java]")    # ignore header file
      temp = line.strip.split("\s")[1]
      if temp.strip.start_with?("<Test>")
        test_name = temp.strip.split("-")[0][6..-1]
        test_result = temp.strip.split("-")[1].eql?("true")? "passed" : "failed"
        tests[test_name] = test_result
        next
      end
      method = temp.strip.split("-")[0]
      visited_times = temp.strip.split("-")[1]
      traces[method] = visited_times     
    end
    Hash[:test => tests, :traces => traces]
  end

  def Helper.method_to_class_coverage(method_coverage_json, output_file)
    traces = method_coverage_json[:traces]
    hash = {}
    traces.each do |key, value|
      class_name, method_name = ""
      if ! key.include?("$")
        method_dot_position = key.count(".")
        method_name = key.split(".")[method_dot_position]
        class_name = key[0..(key.length - method_name.length - 2)]
      else
        class_name = key.split("$")[0]
        method_name = key[(class_name.length + 1)..-1]
        # check if this is reall a class name
        dot_pos = class_name.count(".")
        last_part = class_name.split(".")[dot_pos]
        if last_part.match(/^[a-z]/)
          class_name = class_name[0..(class_name.length - last_part.length - 2)]
        end
        
      end
      new_hash = Hash[class_name => value.strip.to_i]
      hash.merge!(new_hash){|key, old_value, new_value| old_value + new_value}
    end
    sorted_hash = hash.sort_by {|key, value| value}
    f = File.open(output_file, "w")
    sorted_hash.each do |key, value|      
      f.puts("#{key}:#{value}")
    end
    f.close
    sorted_hash
  end
 
  def Helper.convert_coverage_data(input_folder = "/Users/assist/ndataset/dlo/projects/faulttracer-data/lucene37/post-fix/coverage-data/method-cv", output_folder ="/Users/assist/ndataset/dlo/projects/faulttracer-data/lucene37/post-fix/coverage-data/class-cv")
    Dir.entries(input_folder).select{|f| ! f.start_with?(".")}.each do |file|      
      mc_json = Helper.get_method_coverage_data("#{input_folder}/#{file}")
      Helper.method_to_class_coverage(mc_json, "#{output_folder}/#{file}")
    end
  end

  def Helper.clean_method_cv_data(method_cv_dir = "/Users/assist/ndataset/dlo/projects/coverage-data/lucene/method-cv", out = "/Users/assist/ndataset/dlo/projects/coverage-data/lucene/method-cv-ready")
    Dir.entries(method_cv_dir).select{|f| ! f.start_with?(".")}.each do |file|
      mc_json = Helper.get_method_coverage_data("#{method_cv_dir}/#{file}")
      traces = mc_json[:traces]
      f = File.open("#{out}/#{file}", "w")
      traces.each do |key, value|
        f.puts("#{key.split(":")[0]}:#{value}")
      end
      f.close
    end
  end

  ############################################################################
  #
  # Composing final results from queries' results and based on ground truth
  #
  ############################################################################

  def Helper.get_ground_truth(query_id, ground_truth_dir)
    f = File.open(ground_truth_dir + "/" + query_id + ".txt", "r") # remember to add  + ".txt" with the query_id
    locations = f.read().split("\n")
    f.close
    classes = []
    locations.each do |l|
      classes << (l.split("_")[0].include?("$") ? l.split("_")[0].split("$")[0] : l.split("_")[0]).gsub("/", ".") + ".java"
    end
    # puts classes.uniq
    Hash[query_id => classes.uniq]
  end

  def Helper.get_ground_truth_method(query_id, ground_truth_dir)
    f = File.open(ground_truth_dir + "/"+ query_id + ".txt", "r")
    locations = f.read().split("\n")
    f.close
    methods = []
    locations.each do |l|
      prefix = l.split("_")[0].gsub("/", ".").strip
      method_name = l.split("_")[1]

      if method_name.nil?
        puts "[error in get_ground_truth_method] method name is nill"
        puts "[ground truth id] ... [#{query_id}]"
        puts "[line content] ... [#{l}]"
        return        
      end
      
      if method_name.include?("-")
        method_name = method_name.split("-")[0].strip
      elsif method_name.include?("_")
        method_name = method_name.split("_")[0].strip
      end                               
      methods << prefix + "." + method_name
    end      
    Hash[query_id => methods.uniq]   
  end

  def Helper.query_result_to_json(query_id, result_dir)
    f = File.open(result_dir + "/" + query_id, "r")
    lines = f.read().split("\n")
    f.close
    hash = {}
    lines.each do |line|
      key = line.split("\s")[2] # file
      value = line.split("\s")[3] # rank
      hash[key] = value
    end
    Hash[query_id => hash]
  end

  def Helper.query_rank(query_json_result, query_ground_truth, query_id)
    #puts query_ground_truth
    #puts query_id
    locations = query_ground_truth[query_id]
    if locations.nil?
      puts "[error] in query_rank, because query_ground_truth is nil"
      puts  query_ground_truth
      puts query_id
      return
    end
    hash = {}
    locations.each do |l|
      query_json_result[query_id].each do |name, rank|
        if (name.strip.include?(l.strip) && ! l.strip.empty?) || (l.strip.include?(name.strip) && ! name.strip.empty?)
          key = l.strip
          value = rank
          hash[key] = value
          # fix the bug of having the same rank.
          query_json_result[query_id].delete(name) # checked
          break
        end
      end
    end
    Hash[query_id => hash]
  end

  def Helper.compose_results(result_dir = "/Users/assist/dmz/gen-data/ir/class-ir/rhino/all/results/pre-fix", result_name = "rhino.class.subset.slicing", class_or_method = "class", output_file = "/Users/assist/dmz/gen-data/ir/class-ir/rhino/all/results/rhino.class.subset.slicing.json", groundtruth_dir = "/Users/assist/dmz/ir-dataset/ground_truth/rhino", *queries_file)

    #f = File.open(queries_file, "r")
    #queries = f.read().split("\n")
    #f.close

    queries = Dir.entries(result_dir).select {|f| ! f.start_with?(".")}
    
    #     remember to comment this code
    filters = []
    File.readlines("/Users/assist/dmz/gen-data/sIR_and_slicing/out/rhino_found_queries_set.txt").each do |line|
      filters << line.strip.split(".")[0]
    end
    puts filters.size
    queries = filters & queries
    puts queries.size
    #################################

    hash = {}
    queries.each do |query_id|
      puts query_id
      
      q_ground_truth = {}

      if class_or_method.eql?("method")
      then
        q_ground_truth = Helper.get_ground_truth_method(query_id, groundtruth_dir)
                
        if q_ground_truth.nil? || q_ground_truth.empty?
          puts "[error in compose_results]"
          puts query_id
          puts groundtruth_dir
          return
        end        
      elsif class_or_method.eql?("class")
        q_ground_truth = Helper.get_ground_truth(query_id, groundtruth_dir) 
      end

      result_json = Helper.query_result_to_json(query_id, result_dir)
  
      query_ranked = Helper.query_rank(result_json, q_ground_truth, query_id)
      
      if (!query_ranked.nil?) && (!query_ranked[query_id].empty?)
        hash[query_id] = Hash[:results => Hash[query_ranked[query_id].sort_by {|key, value| value.to_i}]]
      end
    end
    
    h = Hash[result_name => hash]

    mrra = 0
    mpa = 0
    
    h[result_name].each do |key, value|
      mrr = 1 / value[:results].values[0].to_f
      value[:MRR] = mrr
      mrra += mrr
      mp = 0
      value[:results].each_with_index do |(name, rank), index|
        mp += (index+1).to_f / rank.to_f 
      end

      mp = mp / value[:results].size 

      value[:MP] = mp
      mpa += mp      
    end

    mrra = mrra.to_f / h[result_name].size
    mpa = mpa.to_f / h[result_name].size

    h[:MRRA] = mrra
    h[:MPA] = mpa
        
    puts h

    Helper.write_json_to_file(h, output_file)
    
    Helper.print_query_results(h, result_name)

    Helper.print_query_results_txtfile(output_file, result_name, "/Users/assist/dmz/gen-data/ir/class-ir/rhino/all/results/rhino.class.subset.slicing.txt")

    Helper.classify_results(output_file, result_name, "/Users/assist/dmz/gen-data/ir/class-ir/rhino/all/results/rhino.class.subset.slicing")

    h
  end

  def Helper.print_query_results(json_result, queries_name)
    top1 = 0
    top5 = 0
    top10 = 0
      
    json_result[queries_name].each_with_index do |(key, value), index|
      puts "-----------------------------------------------------------------------------"
      puts "#{index+1}. #{key} => MRR = #{format("%.2f",value[:MRR])}; MP = #{format("%.2f", value[:MP])}"
      puts "\n"
      value[:results].each do |name, rank|
        puts "#{name} : #{rank}"

        rank = rank.to_i

        if rank == 1
          top1 += 1
          top5 += 1
          top10 += 1
        elsif rank > 1 && rank <= 5
          top5 += 1
          top10 += 1
        elsif rank > 5 && rank <= 10
          top10 += 1
        end
      end
      puts "\n"
    end
    
    puts "\n"
    
    puts "-------------------------------------------------------------------------------"
    
    puts "Overall, Mean MRR = #{format("%.2f", json_result[:MRRA])} and Mean MPA = #{format("%.2f",json_result[:MPA])}"
    puts "Top 1: #{top1}, Top 5: #{top5}, Top 10: #{top10}"
    puts "\n"
  end

  def Helper.classify_results(json_result_file, queries_name, out_text_file)
    
    json_result = Helper.parse_json_file(json_result_file)
    
    cut_offs = [1, 5, 10, 20, 30, 40, 50]
    hash = {}

    cut_offs.each do |ctop|
      count = 0
      ctop = ctop.to_i
      h = {}
      list = []
      json_result[queries_name].each_with_index do |(key, value), index|
        value["results"].each do |name, rank|
          rank = rank.to_i
          if rank <= ctop
            count += 1
            list << key
            break
          end
        end
      end
      h["count"] = count
      h["ids"] = list
      hash[ctop] = h
    end

    all = []
    json_result[queries_name].each do |key, value|
      all << key
    end

    head = hash[cut_offs.max]["ids"]

    #puts all
    #puts head

    tail = all - head
    t = {}
    t["count"] = tail.size
    t["ids"] = tail
    hash[-1] = t

    #puts hash

    f = File.open(out_text_file + "." + cut_offs.max.to_s + ".top", mode="w")
    
    prev_ids = []
    
    hash.each do |key, value|
      if key != -1
        current_ids = value["ids"] - prev_ids
        prev_ids = value["ids"]
          
        f.puts "Top #{key} (#{value["count"]}): #{value["ids"].join(", ")}"
        f.puts "Top #{key} (#{current_ids.size}): #{current_ids.join(", ")}"

      else
        f.puts "The rest (#{value["count"]}): #{value["ids"].join(", ")}"        
      end
      f.puts "\n"
    end

    f.close

    hash    
  end


  def Helper.print_query_results_txtfile(json_result_file, queries_name, out_text_file)

    json_result = Helper.parse_json_file(json_result_file)

    top1 = 0
    top5 = 0
    top10 = 0

    f = File.open(out_text_file, mode="w")
    
    json_result[queries_name].each_with_index do |(key, value), index|
      f.puts "-----------------------------------------------------------------------------"
      f.puts "#{index+1}. #{key} => MRR = #{format("%.2f",value["MRR"])}; MP = #{format("%.2f", value["MP"])}"
      f.puts "\n"
      value["results"].each do |name, rank|
        f.puts "#{name} : #{rank}"

        rank = rank.to_i
        
        if rank == 1
          top1 += 1
          top5 += 1
          top10 += 1
        elsif rank > 1 && rank <= 5
          top5 += 1
          top10 += 1
        elsif rank > 5 && rank <= 10
          top10 += 1
        end
      end
      f.puts "\n"
    end
    
    f.puts "\n"
    
    f.puts "-------------------------------------------------------------------------------"
    
    f.puts "Overall, Mean MRR = #{format("%.2f", json_result["MRRA"])} and Mean MPA = #{format("%.2f",json_result["MPA"])}"
    f.puts "Top 1: #{top1}, Top 5: #{top5}, Top 10: #{top10}"
    f.puts "\n"

    f.close
    
  end

  def Helper.test_mrr_mpa(json_result_file = "/Users/assist/ndataset/dlo/projects/compose-results/ANT-39.json", result_name = "ANT")
    data = Helper.parse_json_file(json_result_file)
    mrra = 0
    mpa = 0
    data[result_name].each do |key, value|      
      mrra += value["MRR"].to_f
      mpa += value["MP"].to_f
    end
    puts mrra / data[result_name].size
    puts mpa / data[result_name].size
  end

  #############################################################################################################

  def Helper.failed_test_clean(test_dir = "/Users/assist/dataset/TCM/DLo/amldata/failed_tests")
    Dir.entries(test_dir).select{|d| ! d.start_with?(".")}.each do |d|
      #puts d
      Dir.foreach(test_dir + "/" + d) do |bug|
        next if bug == "." || bug == ".."
        #puts bug
        tests = []
        File.readlines(test_dir + "/" + d + "/" + bug).each do |line|          
          tests << line.split(":")[0]
        end
        utests = tests.uniq
        if utests.size > 1
          puts d
          puts bug
          puts utests
        end
      end
    end
  end

  def Helper.failed_test_prep(file)
    tests = []
    File.readlines(file).each do |line|          
      tests << line.split(":")[0]
    end
    puts tests.uniq

    
  ###################################################
  #
  # raw slicing data to class-indexed slicing data #
  #
  ###################################################
  end

  def Helper.class_slicing(slice_file, output_file)
    hash = {}
    File.readlines(slice_file).each do |line|
      if line =~ /The dynamic slice for criterion/ || line =~ /Slice consists of/ || line =~ /Computation took/
        next
      end
      method_name = []
      method_name << line.split(".")[line.split(".").size-1].strip
      class_name = line[0..line.strip.size-method_name[0].size-2]
      if class_name.include?("$")
        class_name = class_name.split("$")[0]
      end
      temp = Hash[class_name => method_name]
      hash.merge!(temp){|key, old_value, new_value| old_value + new_value}
    end

    File.open(output_file, "w") do |f|
      hash.each do |key, value|
        f.puts("#{key}:#{value.size}")
      end
    end    
  end

  def Helper.class_slicing_all(slice_dir = "/Users/assist/ndataset/dlo/projects/slices/ant/fixed", output_dir = "/Users/assist/ndataset/dlo/projects/slicing-data/ant/class-slicing")
    Dir.entries(slice_dir).select{|f| (! f.start_with?(".")) && ! (f =~ /err/)}.each do |file|
      puts "processing #{file}"
      Helper.class_slicing(slice_dir + "/" + file, output_dir + "/" + file.split(".")[0] + ".txt")
      puts "#{file} ... [ok]"
    end
  end


  ###################################################
  #
  # raw slicing data to method-indexed slicing data #
  #
  ###################################################

  def Helper.method_slicing(slice_file, output_file)
    hash = {}
    File.readlines(slice_file).each do |line|
      if line =~ /The dynamic slice for criterion/ || line =~ /Slice consists of/ || line =~ /Computation took/
        next
      end      
      method_name = line.split(":")[0].strip
      line_num = []
      line_num << line.split(":")[1].strip
      temp = Hash[method_name => line_num]
      hash.merge!(temp){|key, old_value, new_value| old_value + new_value}
    end

    File.open(output_file, "w") do |f|
      hash.each do |key, value|
        f.puts("#{key}:#{value.size}")
      end
    end    
  end

  def Helper.method_slicing_all(slice_dir = "/Users/assist/ndataset/dlo/projects/slices/ant/fixed", output_dir = "/Users/assist/ndataset/dlo/projects/slicing-data/ant/method-slicing")
    Dir.entries(slice_dir).select{|f| (! f.start_with?(".")) && ! (f =~ /err/)}.each do |file|
      puts "processing #{file}"
      Helper.method_slicing(slice_dir + "/" + file, output_dir + "/" + file.split(".")[0] + ".txt")
      puts "#{file} ... [ok]"
    end
  end


  ############################################ Powerball ##############################################

  def Helper.predict(data_file = "/Users/assist/dmz/ir-scripts/powerball_data")
    data = []
    File.open(data_file, "r") do |f|
      wBalls = []
      f.each_line do |l|
        data.insert(0, l.gsub(/\s+/m, ' ').split(" ")[0..4])
      end
    end

    data.each_with_index do |balls, index|
      puts "#{index}: #{balls.join(" ")}"
    end

    puts "\n"

    max = data.size - 1
    min = 0
    window_size = max + 1

    (window_size .. max+1).to_a.each do |p|
      balls = []
      all = (1..69).to_a
      (p - window_size .. p-1).to_a.each do |i|
        balls =  balls | data[i]
      end

      balls = balls.map(&:to_i)

      predict_balls = (all - balls)

      puts predict_balls

      #puts "#{p}. predict = #{predict_balls.join(" ")} real = #{data[p].join(" ")}  commons =  #{(predict_balls.map(&:to_i) & data[p].map(&:to_i)).join(" ")}"

    end
    
    
  end

end

#Helper.predict

Helper.compose_results
